package io.gitlab.jfronny.inceptum.frontend.cli.commands;

import io.gitlab.jfronny.inceptum.frontend.cli.BaseGitCommand;
import io.gitlab.jfronny.inceptum.frontend.cli.Command;
import io.gitlab.jfronny.inceptum.frontend.cli.CommandArgs;
import io.gitlab.jfronny.inceptum.model.inceptum.InstanceMeta;
import io.gitlab.jfronny.inceptum.util.install.Steps;
import io.gitlab.jfronny.inceptum.util.ConfigHolder;
import io.gitlab.jfronny.inceptum.util.MetaHolder;
import io.gitlab.jfronny.inceptum.util.Utils;
import org.eclipse.jgit.api.Git;
import org.eclipse.jgit.api.errors.GitAPIException;
import org.eclipse.jgit.transport.CredentialsProvider;
import org.eclipse.jgit.transport.URIish;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URISyntaxException;
import java.nio.file.FileAlreadyExistsException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;

public class GitCommand extends Command {
    public GitCommand() {
        super("Aliases to some git commands. Intended to be used in batch commands", "", List.of("git", "sync"), List.of(
                new CommitCommand(),
                new PushCommand(),
                new PullCommand(),
                new ResetCommand(),
                new CloneCommand()
        ));
    }

    @Override
    protected void invoke(CommandArgs args) {
        Utils.LOGGER.error("No known git command was specified");
    }

    private static class CommitCommand extends BaseGitCommand {
        public CommitCommand() {
            super("Adds and commits all changes in the instance.", "[message...]", "commit");
        }

        @Override
        protected void invoke(CommandArgs args, Path instancePath, InstanceMeta meta, Git git, CredentialsProvider credentials) throws GitAPIException {
            git.commit()
                    .setAll(true)
                    .setMessage(args.length > 1
                            ? String.join(" ", args.getArgs())
                            : "Commit at t" + System.currentTimeMillis())
                    .setSign(ConfigHolder.CONFIG.git.signCommits)
                    .setAuthor(ConfigHolder.CONFIG.git.commitUsername, ConfigHolder.CONFIG.git.commitMail)
                    .setCredentialsProvider(credentials)
                    .call();
        }
    }

    private static class PushCommand extends BaseGitCommand {
        public PushCommand() {
            super("Pushes the current branch to the configured remote", "", "push");
        }

        @Override
        protected void invoke(CommandArgs args, Path instancePath, InstanceMeta meta, Git git, CredentialsProvider credentials) throws GitAPIException {
            git.push()
                    .setCredentialsProvider(credentials)
                    .call();
        }
    }

    private static class PullCommand extends BaseGitCommand {
        public PullCommand() {
            super("Pulls remote changes into the current branch and working tree", "", "pull");
        }

        @Override
        protected void invoke(CommandArgs args, Path instancePath, InstanceMeta meta, Git git, CredentialsProvider credentials) throws GitAPIException {
            git.pull()
                    .setRebase(true)
                    .setCredentialsProvider(credentials)
                    .call();
        }
    }

    private static class ResetCommand extends BaseGitCommand {
        public ResetCommand() {
            super("Resets the current instance state to the last commit", "", "reset");
        }

        @Override
        protected void invoke(CommandArgs args, Path instancePath, InstanceMeta meta, Git git, CredentialsProvider credentials) throws GitAPIException {
            git.checkout()
                    .setAllPaths(true)
                    .call();
            git.clean()
                    .setForce(true)
                    .setCleanDirectories(true)
                    .call();
        }
    }

    public static class CloneCommand extends Command {
        public CloneCommand() {
            super("Clones an instance from a URL", "<remote> [name]", "clone");
        }

        @Override
        protected void invoke(CommandArgs args) throws Exception {
            if (args.length == 0) {
                System.err.println("You haven't specified a remote");
                return;
            }
            String name = args.length > 1 ? args.get(1) : getName(args.get(0));
            name = name.replaceAll("^\\.*", "");
            clone(args.get(0), name);
        }

        public static String getName(String url) throws URISyntaxException {
            try {
                String name;
                name = new URIish(url).getHumanishName();
                name = name.replaceAll("^\\.*", "");
                return name;
            }
            catch (IllegalArgumentException e) {
                throw new URISyntaxException(url, "Could not extract name");
            }
        }

        public static void clone(String url, String name) throws GitAPIException, IOException {
            if (!name.matches("^[a-zA-Z0-9]+.*$")) {
                throw new IOException("Invalid name: " + name);
            }
            Path instanceDir = MetaHolder.INSTANCE_DIR.resolve(name);
            if (Files.exists(instanceDir)) {
                throw new FileAlreadyExistsException("An instance by that name already exists");
            }
            Git.cloneRepository()
                    .setURI(url)
                    .setDirectory(instanceDir.toFile())
                    .call()
                    .close();
            Path instanceMeta = instanceDir.resolve("instance.json");
            if (!Files.exists(instanceMeta)) {
                Utils.deleteRecursive(instanceDir);
                throw new FileNotFoundException("Invalid repository: Contains no instance.json");
            }
            Steps.reDownload(instanceDir, Steps.createProcessState());
        }
    }
}
