package io.gitlab.jfronny.inceptum.frontend.gui.window.edit;

import imgui.ImGui;
import imgui.flag.ImGuiInputTextFlags;
import imgui.type.ImString;
import io.gitlab.jfronny.inceptum.Inceptum;
import io.gitlab.jfronny.inceptum.frontend.gui.control.Tab;
import io.gitlab.jfronny.inceptum.frontend.gui.window.GuiUtil;
import io.gitlab.jfronny.inceptum.model.inceptum.Config;
import io.gitlab.jfronny.inceptum.util.ConfigHolder;
import io.gitlab.jfronny.inceptum.util.Utils;
import org.eclipse.jgit.api.Status;
import org.eclipse.jgit.api.errors.GitAPIException;
import org.eclipse.jgit.transport.CredentialsProvider;
import org.eclipse.jgit.transport.UsernamePasswordCredentialsProvider;

import java.io.IOException;

public class GitTab extends Tab {
    private final InstanceEditWindow window;
    private final ImString gitUsername = new ImString(GuiUtil.INPUT_FIELD_LENGTH);
    private final ImString gitPassword = new ImString(GuiUtil.INPUT_FIELD_LENGTH);
    protected CredentialsProvider credentialsProvider;

    public GitTab(InstanceEditWindow window) {
        super("Git");
        this.window = window;
        if (ConfigHolder.CONFIG.git.instanceAuths.containsKey(window.name)) {
            Config.GitConfig.GitAuth config = ConfigHolder.CONFIG.git.instanceAuths.get(window.name);
            gitUsername.set(config.username);
            gitPassword.set(config.password);
            credentialsProvider = new UsernamePasswordCredentialsProvider(config.username, config.password);
        }
        else credentialsProvider = CredentialsProvider.getDefault();
    }

    @Override
    protected void renderInner() {
        Status status = null;
        try {
            status = window.git.status().call();
        } catch (GitAPIException e) {
            Utils.LOGGER.error("Could not check repo status", e);
        }

        if (status != null && status.isClean()) {
            ImGui.text("There are un-committed changes");
            for (String change : status.getUncommittedChanges()) {
                ImGui.bulletText(change);
            }
            for (String s : status.getUntracked()) {
                ImGui.bulletText(s);
            }
            if (ImGui.button("Commit")) {
                try {
                    Inceptum.getInput("Commit message", "Commit at t" + System.currentTimeMillis(), message -> {
                        try {
                            window.git.commit()
                                    .setAll(true)
                                    .setMessage(message)
                                    .setSign(ConfigHolder.CONFIG.git.signCommits)
                                    .setAuthor(ConfigHolder.CONFIG.git.commitUsername, ConfigHolder.CONFIG.git.commitMail)
                                    .setCredentialsProvider(credentialsProvider)
                                    .call();
                        } catch (GitAPIException e) {
                            Inceptum.showError("Could not commit", e);
                        }
                    },() -> {});
                } catch (IOException e) {
                    Inceptum.showError("Could not show dialog", e);
                }
            }
            ImGui.sameLine();
        }
        if (ImGui.button("Push")) {
            try {
                window.git.push()
                        .setCredentialsProvider(credentialsProvider)
                        .call();
            } catch (GitAPIException e) {
                Inceptum.showError("Could not push", e);
            }
        }
        ImGui.sameLine();
        if (ImGui.button("Pull")) {
            try {
                window.git.pull()
                        .setRebase(true)
                        .setCredentialsProvider(credentialsProvider)
                        .call();
            } catch (GitAPIException e) {
                Inceptum.showError("Could not pull", e);
            }
        }
        ImGui.sameLine();
        if (ImGui.button("Reset")) {
            Inceptum.showOkCancel("This will erase all changes since the last commit.\nAre you sure?", "Are you sure?", () -> {
                try {
                    window.git.checkout()
                            .setAllPaths(true)
                            .call();
                    window.git.clean()
                            .setForce(true)
                            .setCleanDirectories(true)
                            .call();
                } catch (GitAPIException e) {
                    Inceptum.showError("Could not reset", e);
                }
            }, () -> {});
        }
        ImGui.text("Authentication");
        boolean update = ImGui.inputText("Username", gitUsername);
        update |= ImGui.inputText("Password", gitPassword, ImGuiInputTextFlags.Password);
        if (update) {
            if (!ConfigHolder.CONFIG.git.instanceAuths.containsKey(window.name))
                ConfigHolder.CONFIG.git.instanceAuths.put(window.name, new Config.GitConfig.GitAuth());
            Config.GitConfig.GitAuth config = ConfigHolder.CONFIG.git.instanceAuths.get(window.name);
            config.username = gitUsername.get();
            config.password = gitPassword.get();
            ConfigHolder.saveConfig();
            credentialsProvider = new UsernamePasswordCredentialsProvider(config.username, config.password);
        }
        ImGui.text("Please note that Inceptum's git integration intentionally does not cover advanced usage");
        ImGui.text("Using software intended to work with git (such as the git CLI or GitHub Desktop) is recommended");
    }

    @Override
    protected boolean canShow() {
        return window.git != null;
    }
}
