package io.gitlab.jfronny.inceptum.frontend.gui.window.edit;

import imgui.ImGui;
import imgui.type.ImBoolean;
import io.gitlab.jfronny.inceptum.Inceptum;
import io.gitlab.jfronny.inceptum.InceptumGui;
import io.gitlab.jfronny.inceptum.frontend.gui.control.Tab;
import io.gitlab.jfronny.inceptum.frontend.gui.window.AddModWindow;
import io.gitlab.jfronny.inceptum.util.ModManager;
import io.gitlab.jfronny.inceptum.util.PathUtil;
import io.gitlab.jfronny.inceptum.util.Utils;
import io.gitlab.jfronny.inceptum.util.mds.IWModDescription;
import io.gitlab.jfronny.inceptum.util.source.ModSource;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

public class ModsTab extends Tab {
    private final InstanceEditWindow window;
    private final ImBoolean filterUpdates = new ImBoolean();
    private Path selected = null;

    public ModsTab(InstanceEditWindow window) {
        super("Mods");
        this.window = window;
    }

    @Override
    protected void renderInner() {
        window.lastTabWasMods = true;
        if (!Files.exists(window.path.resolve("mods"))) {
            try {
                Files.createDirectories(window.path.resolve("mods"));
            } catch (IOException e) {
                Utils.LOGGER.error("Could not create mods directory which was missing from this modded instance", e);
            }
        }
        ImGui.beginChild("mods select", 200, 0);
        if (ImGui.button("Add")) {
            InceptumGui.WINDOWS.add(new AddModWindow(window.path.resolve("mods"), window.instance, window.mds));
        }
        ImGui.sameLine();
        if (Files.exists(window.path.resolve("mods")) && ImGui.button("Show")) {
            Utils.openFile(window.path.resolve("mods").toFile());
        }
        ImGui.sameLine();
        if (Files.exists(window.path.resolve("config")) && ImGui.button("Configs")) {
            Utils.openFile(window.path.resolve("config").toFile());
        }
        try {
            Set<IWModDescription> modSet = window.mds.getMods();
            boolean updatesFound = false;
            float scannedPercentage = 0;
            boolean hasUnScanned = false;
            for (IWModDescription mod : modSet) {
                if (window.mds.hasScanned(mod.path())) scannedPercentage++;
                else hasUnScanned = true;
                if (mod.mod().isEmpty()) continue;
                for (Optional<ModSource> value : mod.mod().get().sources.values()) {
                    if (value.isPresent()) {
                        updatesFound = true;
                        break;
                    }
                    if (updatesFound)
                        break;
                }
            }
            scannedPercentage /= modSet.size();
            if (hasUnScanned) ImGui.progressBar(scannedPercentage);
            if (updatesFound)
                ImGui.checkbox("Updatable", filterUpdates);
            else
                filterUpdates.set(false);
            ImGui.separator();
            for (IWModDescription mod : modSet) {
                updatesFound = false;
                if (mod.mod().isPresent()) {
                    for (Optional<ModSource> value : mod.mod().get().sources.values()) {
                        updatesFound |= value.isPresent();
                    }
                }
                if (filterUpdates.get() && !updatesFound) continue;
                boolean wasEnabled = PathUtil.isEnabled(mod.path());
                if (ImGui.checkbox("##" + mod.getName(), wasEnabled)) {
                    Path newSel = PathUtil.toggle(mod.path());
                    try {
                        Files.move(mod.path(), newSel);
                        if (mod.path().equals(selected)) selected = newSel;
                        mod = new IWModDescription(newSel, mod);
                    } catch (IOException e) {
                        Inceptum.showError("Could not change disabled state", e);
                    }
                }
                ImGui.sameLine();
                if (ImGui.button(mod.getName())) selected = mod.path();
            }
        } catch (IOException e) {
            Utils.LOGGER.error("Could not show mod list", e);
        }
        ImGui.endChild();
        ImGui.sameLine();
        ImGui.beginGroup();
        if (selected == null) {
            ImGui.text("Select a mod to view settings");
        } else if (window.mds.hasScanned(selected)) {
            IWModDescription md = window.mds.get(selected);
            ImGui.text(md.getName());
            ImGui.separator();
            for (String s : md.getDescription()) {
                ImGui.text(s);
            }
            ImGui.separator();
            if (md.mod().isPresent()) {
                Map<ModSource, Optional<ModSource>> sources = md.mod().get().sources;
                ImGui.text("Sources:");
                if (sources.isEmpty())
                    ImGui.bulletText("Local Drive");
                else {
                    for (Map.Entry<ModSource, Optional<ModSource>> source : sources.entrySet()) {
                        ImGui.bulletText(source.getKey().getName());
                        Optional<ModSource> ms = source.getValue();
                        if (ms.isPresent()) {
                            ImGui.sameLine();
                            if (ImGui.button("Update to " + ms.get().getVersion())) {
                                try {
                                    Path imodPath = md.imod().isPresent() ? md.imod().get() : window.path.resolve("mods").resolve(ms.get().getShortName() + PathUtil.EXT_IMOD);
                                    AddModWindow.DownloadMeta dm = AddModWindow.download(ms.get(), imodPath, window.mds);
                                    Files.delete(md.path());
                                    if (md.imod().isPresent() && Files.exists(md.imod().get()))
                                        Files.delete(md.imod().get());
                                    dm.write();
                                    selected = imodPath;
                                    window.mds.invalidate(imodPath);
                                } catch (IOException e) {
                                    Inceptum.showError("Update failed", e);
                                }
                            }
                        }
                    }
                }
            }
            if (ImGui.button("Delete")) {
                if (md.mod().isPresent() && !md.mod().get().dependents.isEmpty())
                    Inceptum.showError("This mod still has the following dependent mods installed: " + String.join(", ", md.mod().get().dependents), "Dependents present");
                else {
                    try {
                        delete(md);
                    } catch (IOException e) {
                        Inceptum.showError("Couldn't delete the file", e);
                    }
                    selected = null;
                }
            }
        }
        else {
            ImGui.text("This mod has not yet been scanned, please be patient");
        }
        ImGui.endGroup();
    }

    private void delete(IWModDescription md) throws IOException {
        ModManager.delete(md, window.path.resolve("mods"), window.mds);
    }

    @Override
    protected boolean canShow() {
        return window.instance.isFabric();
    }
}
