package io.gitlab.jfronny.inceptum.util.mds;

import io.gitlab.jfronny.inceptum.model.inceptum.InstanceMeta;
import io.gitlab.jfronny.inceptum.util.PathUtil;
import io.gitlab.jfronny.inceptum.util.Utils;

import java.io.IOException;
import java.nio.file.*;
import java.util.*;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.function.BiConsumer;

import static java.nio.file.StandardWatchEventKinds.*;

class ModsDirScannerImpl implements ModsDirScanner {
    private static final Map<Path, ModsDirScannerImpl> SCANNERS = new HashMap<>();
    private static final ExecutorService POOL = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors(), new NamedThreadFactory("mds"));
    private final Map<Path, IWModDescription> descriptions = new HashMap<>();
    private final Set<Path> scannedPaths = new HashSet<>();
    private final Thread th;
    private final Path modsDir;
    private final InstanceMeta instance;
    private final WatchService service;
    private boolean disposed = false;

    private ModsDirScannerImpl(Path modsDir, InstanceMeta instance) throws IOException {
        this.modsDir = modsDir;
        this.instance = instance;
        this.th = new Thread(this::scanTaskInternal, "mds-" + modsDir.getParent().getFileName());
        this.service = FileSystems.getDefault().newWatchService();
        modsDir.register(service, ENTRY_MODIFY, ENTRY_CREATE, ENTRY_DELETE);
    }

    protected static ModsDirScannerImpl get(Path modsDir, InstanceMeta instance) throws IOException {
        if (SCANNERS.containsKey(modsDir)) {
            ModsDirScannerImpl mds = SCANNERS.get(modsDir);
            if (mds.instance.equals(instance))
                return mds;
            mds.close();
        }
        ModsDirScannerImpl mds = new ModsDirScannerImpl(modsDir, instance);
        SCANNERS.put(modsDir, mds);
        return mds;
    }

    @Override
    public boolean isComplete() {
        if (!Files.isDirectory(modsDir)) return true;
        try {
            for (Path path : Utils.ls(modsDir)) {
                if (!descriptions.containsKey(path)) return false;
            }
        } catch (IOException e) {
            Utils.LOGGER.error("Could not list files in mods dir", e);
        }
        return true;
    }

    @Override
    public void start() {
        if (!th.isAlive())
            th.start();
    }

    @Override
    public String getGameVersion() {
        return instance.getMinecraftVersion();
    }

    @Override
    public Set<IWModDescription> getMods() throws IOException {
        Set<IWModDescription> mods = new TreeSet<>();
        if (Files.isDirectory(modsDir)) {
            for (Path path : Utils.ls(modsDir)) {
                if (PathUtil.isImod(path) && Files.exists(PathUtil.trimImod(path)))
                    continue;
                mods.add(get(path));
            }
        }
        return mods;
    }

    @Override
    public IWModDescription get(Path path) {
        if (!descriptions.containsKey(path)) {
            // not yet scanned
            descriptions.put(path, new IWModDescription(path));
        }
        return descriptions.get(path);
    }

    @Override
    public void invalidate(Path path) {
        descriptions.remove(path);
        scannedPaths.remove(path);
    }

    public boolean hasScanned(Path path) {
        return scannedPaths.contains(path)
                || scannedPaths.contains(path.getParent().resolve(path.getFileName().toString() + ".imod"));
    }

    private void scanTaskInternal() {
        while (!disposed) {
            runOnce((path, iwModDescription) -> {});
        }
    }

    @Override
    public void runOnce(BiConsumer<Path, IWModDescription> discovered) {
        try {
            if (!Files.isDirectory(modsDir)) {
                return;
            }
            Set<Future<?>> tasks = new HashSet<>();
            String mc = instance.getMinecraftVersion();
            discovered = discovered.andThen((path, iwModDescription) -> {
                scannedPaths.add(path);
                descriptions.put(path, iwModDescription);
            });
            Set<Path> toScan;
            if (descriptions.isEmpty()) {
                toScan = Set.copyOf(Utils.ls(modsDir));
            }
            else {
                toScan = new HashSet<>();
                WatchKey key = service.poll();
                if (key != null) {
                    for (WatchEvent<?> event : key.pollEvents()) {
                        if (event.context() instanceof Path p)
                            toScan.add(p);
                    }
                    key.reset();
                }
                Utils.ls(modsDir, path -> {
                    if (!descriptions.containsKey(path))
                        toScan.add(path);
                });
            }
            for (Path p : toScan) {
                tasks.add(POOL.submit(new FileScanTask(p, discovered, mc)));
            }
            for (Future<?> task : tasks) {
                try {
                    task.get();
                } catch (ExecutionException e) {
                    Utils.LOGGER.error("Failed to execute ScanTask", e);
                }
            }
        } catch (IOException | InterruptedException e) {
            Utils.LOGGER.error("Could not list mods", e);
        }
    }

    public static void closeAll() {
        for (ModsDirScannerImpl value : SCANNERS.values().toArray(ModsDirScannerImpl[]::new)) {
            value.close();
        }
        POOL.shutdown();
    }

    @Override
    public void close() {
        disposed = true;
        SCANNERS.remove(modsDir);
    }
}
