package io.gitlab.jfronny.inceptum.util.source;

import io.gitlab.jfronny.commons.HashUtils;
import io.gitlab.jfronny.commons.tuple.Tuple;
import io.gitlab.jfronny.inceptum.model.modrinth.ModrinthProject;
import io.gitlab.jfronny.inceptum.model.modrinth.ModrinthVersion;
import io.gitlab.jfronny.inceptum.util.Utils;
import io.gitlab.jfronny.inceptum.util.api.ModrinthApi;
import io.gitlab.jfronny.inceptum.util.cache.MemoryCache;

import java.io.IOException;
import java.net.URISyntaxException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.HashSet;
import java.util.Optional;
import java.util.Set;

public final class ModrinthModSource implements ModSource {
    private static final MemoryCache<Tuple<String, String>, Optional<ModSource>> UPDATE_CACHE = new MemoryCache<>();
    private static final MemoryCache<String, Set<ModSource>> DEPENDENCIES_CACHE = new MemoryCache<>();

    private final String versionId;
    private final ModrinthVersion current;
    private final ModrinthProject mod;

    public ModrinthModSource(String versionId) throws IOException {
        this.versionId = versionId;
        this.current = ModrinthApi.getVersion(versionId);
        this.mod = ModrinthApi.getMod(getModId());
    }

    @Override
    public ModDownload download() throws IOException {
        ModrinthVersion.File file = current.files.get(0);
        Path path = getJarPath();
        try {
            Utils.downloadFile(file.url, file.hashes.sha1, path);
        } catch (URISyntaxException e) {
            throw new IOException("Could not download file", e);
        }
        return new ModDownload(file.hashes.sha1, HashUtils.murmur2(Files.readAllBytes(path)), path);
    }

    @Override
    public Set<ModSource> getDependencies(String gameVersion) throws IOException {
        return DEPENDENCIES_CACHE.get(versionId, () -> {
            Set<ModSource> deps = new HashSet<>();
            for (ModrinthVersion.Dependency dependency : current.dependencies) {
                //TODO show optional dependencies
                if (dependency.dependency_type == ModrinthVersion.Dependency.DependencyType.required)
                    deps.add(new ModrinthModSource(dependency.version_id));
            }
            return Set.copyOf(deps);
        });
    }

    @Override
    public Optional<ModSource> getUpdate(String gameVersion) throws IOException {
        return UPDATE_CACHE.get(Tuple.of(versionId, gameVersion), () -> {
            ModrinthVersion stable = null;
            ModrinthVersion beta = null;
            ModrinthVersion latest = null;
            for (ModrinthVersion version : ModrinthApi.getVersions(getModId())) {
                if (version.game_versions.contains(gameVersion) && version.loaders.contains("fabric")) {
                    latest = version;
                    if (version.version_type == ModrinthVersion.VersionType.beta || version.version_type == ModrinthVersion.VersionType.release)
                        beta = version;
                    if (version.version_type == ModrinthVersion.VersionType.release)
                        stable = version;
                }
            }
            ModrinthVersion next = switch (current.version_type) {
                case alpha -> latest;
                case beta -> beta;
                case release -> stable;
            };
            if (next == null) return Optional.empty();
            if (next.version_number.equals(current.version_number)) return Optional.empty();
            return Optional.of(new ModrinthModSource(next.id));
        });
    }

    @Override
    public String getVersion() {
        return current.version_number;
    }

    @Override
    public String getName() {
        return "modrinth/" + getShortName() + '/' + current.version_number;
    }

    @Override
    public String getShortName() {
        return (mod.slug == null ? mod.id : mod.slug);
    }

    @Override
    public String getFileName() {
        return current.files.get(0).filename;
    }

    @Override
    public boolean equals(ModSource other) {
        return other instanceof ModrinthModSource ms && ms.getModId().equals(getModId()) && ms.versionId.equals(versionId);
    }

    public String getVersionId() {
        return versionId;
    }

    public String getModId() {
        return current.project_id;
    }
}
